! WHIZARD 1.10 Apr 20 2001
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module beams

  use kinds, only: default !NODEP!
  use file_utils !NODEP!
  use mpi90, only: mpi90_broadcast !NODEP!
  use constants
  use parameters !NODEP!
!  use polarization
  use particles
  use events
  use pdflib_interface !NODEP!
  use limits, only: KIREPS

  implicit none
  private

  type, public :: beam_parameter_set
     real(kind=default), dimension(3) :: polarization
!     type(spin_density_vector) :: rho
!     logical :: rho_set
     logical :: fixed_energy
     integer :: n_strfun
     logical :: CIRCE_on, CIRCE_map
     real(kind=default) :: CIRCE_beta, CIRCE_gamma
     real(kind=default) :: CIRCE_sqrts
     integer :: CIRCE_prt, CIRCE_ver, CIRCE_rev, CIRCE_acc, CIRCE_chat
     logical :: ISR_on
     real(kind=default) :: ISR_alpha, ISR_m_in, ISR_sqrts
     real(kind=default) :: ISR_power, ISR_x0
     logical :: EPA_on, EPA_map
     integer :: EPA_in_prt 
     real(kind=default) :: EPA_alpha, EPA_mX, EPA_Q_max, EPA_x0, EPA_x1
     real(kind=default) :: EPA_msq, EPA_s, EPA_c0, EPA_c1
     logical :: PDF_on
     integer :: PDF_in_prt, PDF_ngroup, PDF_nset, PDF_nfl, PDF_lo
     logical :: PDF_running_scale
     real(kind=default) :: PDF_scale, PDF_QCDL4, PDF_mtop
  end type beam_parameter_set

  public :: create, destroy
  public :: set
  public :: read, write
  public :: mpi90_broadcast
  public :: structure_function

  interface create
     module procedure beam_create_single
     module procedure beam_create_multi
  end interface
  interface destroy
     module procedure beam_destroy_single
     module procedure beam_destroy_multi
  end interface
  interface set
     module procedure beam_par_set
  end interface
  interface read
     module procedure read_beam_parameters_unit
     module procedure read_beam_parameters_name
  end interface
  interface write
     module procedure write_beam_parameters_unit
     module procedure write_beam_parameters_name
  end interface
  interface mpi90_broadcast
     module procedure beam_par_broadcast
  end interface

contains

  subroutine beam_create_multi(beam)
    type(beam_parameter_set), dimension(:), intent(out) :: beam
    integer :: i
    do i=1, size(beam)
       call beam_create_single(beam(i))
    end do
  end subroutine beam_create_multi
  subroutine beam_create_single(beam)
    type(beam_parameter_set), intent(out) :: beam
    beam%polarization = 0
!    beam%rho_set = .false.
    beam%fixed_energy = .true.
    beam%n_strfun = 0
    beam%CIRCE_on = .false.
    beam%CIRCE_map = .true.
    beam%CIRCE_beta = 0
    beam%CIRCE_gamma = 0
    beam%CIRCE_sqrts = 0
    beam%CIRCE_prt = 0
    beam%CIRCE_ver = 0
    beam%CIRCE_rev = 0
    beam%CIRCE_acc = 0
    beam%CIRCE_chat = 0
    beam%ISR_on = .false.
    beam%ISR_alpha = 0
    beam%ISR_m_in = 0
    beam%ISR_sqrts = 0
    beam%ISR_power = 0
    beam%ISR_x0 = 0
    beam%EPA_on = .false.
    beam%EPA_map = .true.
    beam%EPA_in_prt = 0
    beam%EPA_alpha = 0
    beam%EPA_mX = 0
    beam%EPA_Q_max = 0
    beam%EPA_x0 = 0
    beam%EPA_x1 = 0
    beam%EPA_msq = 0
    beam%EPA_s = 0
    beam%EPA_c0 = 0
    beam%EPA_c1 = 0
    beam%PDF_on = .false.
    beam%PDF_in_prt = 0
    beam%PDF_ngroup = 0
    beam%PDF_nset = 0
    beam%PDF_nfl = 0
    beam%PDF_lo = 0
    beam%PDF_running_scale = .false.
    beam%PDF_scale = 0
    beam%PDF_QCDL4 = 0
    beam%PDF_mtop = 0
  end subroutine beam_create_single

  subroutine beam_destroy_multi(beam)
    type(beam_parameter_set), dimension(:), intent(inout) :: beam
    integer :: i
    do i=1, size(beam)
       call beam_destroy_single(beam(i))
    end do
  end subroutine beam_destroy_multi
  subroutine beam_destroy_single(beam)
    type(beam_parameter_set), intent(inout) :: beam
!     if (beam%rho_set) then
!        call destroy(beam%rho)
!        beam%rho_set = .false.
!     end if
  end subroutine beam_destroy_single

  subroutine beam_par_set(beam, par, code, sqrts, polarized, beam_user)
    type(beam_parameter_set), dimension(2), intent(out) :: beam
    type(parameter_set), intent(in) :: par
    integer, dimension(:), intent(in) :: code
    real(kind=default), intent(in) :: sqrts
    logical, intent(in) :: polarized
    type(beam_parameter_set), dimension(2), intent(in), optional :: beam_user
    integer :: i
    do i=1,2
!        call create(beam(i)%rho, &
!             &      (/ particle_spin_type(code(i)) /) )
!        beam(i)%rho_set = .true.
       beam(i)%polarization = 0
       beam(i)%fixed_energy = .true.
       beam(i)%n_strfun = 0
       if (present(beam_user)) then
          if (polarized) beam(i)%polarization = beam_user(i)%polarization
          if (.not.beam_user(i)%fixed_energy) then
             if (beam_user(i)%CIRCE_on .and. &
                  & (code(i)==PHOTON .or. abs(code(i))==ELECTRON)) then
                beam(i)%fixed_energy = .false.
                beam(i)%n_strfun = beam(i)%n_strfun+1
                beam(i)%CIRCE_on = .true.
                beam(i)%CIRCE_map = beam_user(i)%CIRCE_map
                beam(i)%CIRCE_sqrts = beam_user(i)%CIRCE_sqrts
                if (beam(i)%CIRCE_sqrts ==0) &
                     & beam(i)%CIRCE_sqrts = sqrts
                if (beam(i)%CIRCE_sqrts<=0) &
                     & stop 'CIRCE needs positive scale CIRCE_sqrts'
                beam(i)%CIRCE_prt = code(i)
                beam(i)%CIRCE_acc = beam_user(i)%CIRCE_acc
                beam(i)%CIRCE_ver = beam_user(i)%CIRCE_ver
                beam(i)%CIRCE_rev = beam_user(i)%CIRCE_rev
                beam(i)%CIRCE_chat = beam_user(i)%CIRCE_chat
             end if
             if (i==2 .and. any(beam%CIRCE_on)) then
                if (.not.all(beam%CIRCE_on)) &
                     & stop "Beamstrahlung must be defined for both beams, if any"
                if (beam(1)%CIRCE_sqrts /= beam(2)%CIRCE_sqrts &
                     & .or. beam(1)%CIRCE_ver /= beam(2)%CIRCE_ver &
                     & .or. beam(1)%CIRCE_rev /= beam(2)%CIRCE_rev &
                     & .or. beam(1)%CIRCE_acc /= beam(2)%CIRCE_acc &
                     & .or. beam(1)%CIRCE_chat/= beam(2)%CIRCE_chat) &
                     & stop "CIRCE parameters must coincide for both beams"
                call circes (0.d0, 0.d0, dble(beam(1)%CIRCE_sqrts), &
                     & beam(1)%CIRCE_acc, &
                     & beam(1)%CIRCE_ver, beam(1)%CIRCE_rev, beam(1)%CIRCE_chat)
                call map_CIRCE_set_powers (beam(1)%CIRCE_beta, beam(1)%CIRCE_gamma, code(1))
                call map_CIRCE_set_powers (beam(2)%CIRCE_beta, beam(2)%CIRCE_gamma, code(2))
             end if
             if (beam_user(i)%ISR_on .and. particle_charge(code(i))/=0) then
                beam(i)%fixed_energy = .false.
                beam(i)%n_strfun = beam(i)%n_strfun + 1
                beam(i)%ISR_on = .true.
                beam(i)%ISR_alpha = beam_user(i)%ISR_alpha
                if (beam(i)%ISR_alpha ==0) &
                     & beam(i)%ISR_alpha = par%EE**2/(4*pi)
                if (beam(i)%ISR_alpha <=0) &
                     & stop 'ISR needs positive coupling ISR_alpha'
                beam(i)%ISR_m_in  = beam_user(i)%ISR_m_in
                if (beam(i)%ISR_m_in  ==0) &
                     & beam(i)%ISR_m_in  = particle_mass(code(i), par)
                if (beam(i)%ISR_m_in <=0) &
                     & stop 'ISR needs positive mass ISR_m_in'
                beam(i)%ISR_sqrts = beam_user(i)%ISR_sqrts
                if (beam(i)%ISR_sqrts ==0) &
                     & beam(i)%ISR_sqrts = sqrts
                if (beam(i)%ISR_sqrts<=0) &
                     & stop 'ISR needs positive scale ISR_sqrts'
                if (beam(i)%ISR_sqrts<=beam(i)%ISR_m_in) &
                     & stop 'ISR_m_in must be smaller than ISR_sqrts'
                beam(i)%ISR_power = &
                     & 2/(beam(i)%ISR_alpha/pi &
                     &    * particle_charge(code(i))**2 &
                     &    * log(beam(i)%ISR_sqrts / beam(i)%ISR_m_in))
                beam(i)%ISR_x0 = 1 - tiny(1._default)**(1/beam(i)%ISR_power)
                if (beam(i)%ISR_power<=1) stop 'ISR power out of range'
             end if
             if (beam_user(i)%EPA_on .and. abs(code(i))==PHOTON) then
                beam(i)%fixed_energy = .false.
                beam(i)%n_strfun = beam(i)%n_strfun + 1
                beam(i)%EPA_on = .true.
                beam(i)%EPA_map = beam_user(i)%EPA_map
                beam(i)%EPA_in_prt = beam_user(i)%EPA_in_prt
                if (beam(i)%EPA_in_prt == 0) &
                     & stop 'EPA_in_prt (PDG code) must be specified.'
                beam(i)%EPA_msq = particle_mass(beam(i)%EPA_in_prt, par)**2
                if (beam(i)%EPA_msq == 0) &
                     & stop 'EPA_in_prt must have nonzero mass.'
                if (particle_charge(beam(i)%EPA_in_prt) == 0) &
                     & stop 'EPA_in_prt must have nozero charge.'
                beam(i)%EPA_alpha = beam_user(i)%EPA_alpha
                if (beam(i)%EPA_alpha == 0) &
                     & beam(i)%EPA_alpha = par%EE**2/(4*pi)
                if (beam(i)%EPA_alpha <= 0) &
                     & stop 'EPA needs positive coupling EPA_alpha'
                if (beam(i)%EPA_map) then
                   beam(i)%EPA_mX  = beam_user(i)%EPA_mX
                   ! if (beam(i)%EPA_mX == 0) then
                   !    do j=1, evt%n_out
                   !       beam(i)%EPA_mX = beam(i)%EPA_mX + particle_mass(code(evt%prt(j))
                   !    end do
                   ! end if
                   if (beam(i)%EPA_mX <= 0) &
                        & stop 'EPA needs positive mass EPA_mX'
                   if (beam(i)%EPA_mX >= sqrts) &
                        & stop 'EPA_mX must be smaller than sqrt(s)'
                end if
                beam(i)%EPA_Q_max = beam_user(i)%EPA_Q_max
                if (beam(i)%EPA_Q_max <=0) &
                     & stop 'EPA needs positive cutoff EPA_Q_max'
                if (beam(i)%EPA_Q_max >= sqrts) &
                     & stop 'EPA_Q_max must be smaller than sqrt(s)'
                beam(i)%EPA_x0 = beam_user(i)%EPA_x0
                if (beam(i)%EPA_map) then
                   if (beam(i)%EPA_x0 < (beam(i)%EPA_mX / sqrts)**2) &
                        & beam(i)%EPA_x0 = (beam(i)%EPA_mX / sqrts)**2
                else
                   if (beam(i)%EPA_x0 < 0) beam(i)%EPA_x0 = 0
                end if
                beam(i)%EPA_x1 = beam_user(i)%EPA_x1
                if (beam(i)%EPA_x1 == 0) beam(i)%EPA_x1 = 1
                if (beam(i)%EPA_x1 > 1) &
                     & stop 'EPA_x1 must not be larger than 1'
                if (beam(i)%EPA_x1 <= beam(i)%EPA_x0) &
                     & stop 'EPA_x0 must be smaller than EPA_x1'
                beam(i)%EPA_s = sqrts**2
                if (beam(i)%EPA_map) then
                   beam(i)%EPA_c0 = beam(i)%EPA_alpha/pi &
                        & * particle_charge(beam(i)%EPA_in_prt)**2 &
                        & * log(beam(i)%EPA_x0) &
                        & * log(beam_user(i)%EPA_Q_max**2 / beam(i)%EPA_msq &
                        &       / beam(i)%EPA_x0)
                   beam(i)%EPA_c1 = beam(i)%EPA_alpha/pi &
                        & * particle_charge(beam(i)%EPA_in_prt)**2 &
                        & * log(beam(i)%EPA_x1) &
                        & * log(beam_user(i)%EPA_Q_max**2 / beam(i)%EPA_msq &
                        &       / beam(i)%EPA_x1)
                   if (beam(i)%EPA_c1 <= beam(i)%EPA_c0) &
                        & stop 'EPA_Q_max is too small'
                else
                   beam(i)%EPA_c0 = beam(i)%EPA_alpha/pi &
                        & * particle_charge(beam(i)%EPA_in_prt)**2 &
                        & * beam(i)%EPA_x0
                   beam(i)%EPA_c1 = beam(i)%EPA_alpha/pi &
                        & * particle_charge(beam(i)%EPA_in_prt)**2 &
                        & * beam(i)%EPA_x1
                end if
             end if
             if (beam(i)%EPA_on .and. beam(i)%CIRCE_on) &
                  & stop "EPA and beamstrahlung must not simultaneously be switched on" 
             if (beam_user(i)%PDF_on) then
                select case(abs(code(i)))
                case (D_QUARK:T_QUARK,GLUON,GLUON2)
                   beam(i)%fixed_energy = .false.
                   beam(i)%n_strfun = beam(i)%n_strfun + 1
                   beam(i)%PDF_on = .true.
                   select case (beam_user(i)%PDF_in_prt)
                   case (PROTON, -PROTON, PHOTON)
                      beam(i)%PDF_in_prt = beam_user(i)%PDF_in_prt
                      beam(i)%PDF_ngroup = beam_user(i)%PDF_ngroup
                      beam(i)%PDF_nset = beam_user(i)%PDF_nset
                      beam(i)%PDF_nfl = beam_user(i)%PDF_nfl
                      beam(i)%PDF_lo = beam_user(i)%PDF_lo
                      beam(i)%PDF_QCDL4 = beam_user(i)%PDF_QCDL4
                      beam(i)%PDF_mtop = par%mtop
                      beam(i)%PDF_running_scale = beam_user(i)%PDF_running_scale
                      if (.not.beam(i)%PDF_running_scale) then
                         beam(i)%PDF_scale = beam_user(i)%PDF_scale
                         if (beam(i)%PDF_scale <= 0) &
                              stop "PDF scale must either defined as running or given a positive value"
                      end if
                   case default
                      stop 'Only proton, antiproton and photon allowed as PDF initial particle'
                   end select
                end select
             end if
             if (i==2 .and. any(beam%PDF_on)) then
                if (all(beam%PDF_on)) then
                   if (abs(beam(1)%PDF_in_prt) == abs(beam(2)%PDF_in_prt) &
                        & .and. (beam(1)%PDF_ngroup /= beam(2)%PDF_ngroup &
                        &   .or. beam(1)%PDF_nset /= beam(2)%PDF_nset &
                        &   .or. beam(1)%PDF_nfl /= beam(2)%PDF_nfl &
                        &   .or. beam(1)%PDF_lo /= beam(2)%PDF_lo &
                        &   .or. beam(1)%PDF_QCDL4 /= beam(2)%PDF_QCDL4 &
                        &   .or. beam(1)%PDF_scale /= beam(2)%PDF_scale)) &
                        & stop "PDF parameters must coincide for both beams"
                end if
                if (beam(1)%PDF_on) then
                   call pdflib_init (beam(1)%PDF_in_prt, beam(1)%PDF_ngroup, &
                        &            beam(1)%PDF_nset, beam(1)%PDF_nfl, beam(1)%PDF_lo, &
                        &            beam(1)%PDF_QCDL4, beam(1)%PDF_mtop, first=.true.)
                else
                   call pdflib_init (beam(2)%PDF_in_prt, beam(2)%PDF_ngroup, &
                        &            beam(2)%PDF_nset, beam(2)%PDF_nfl, beam(2)%PDF_lo, &
                        &            beam(2)%PDF_QCDL4, beam(2)%PDF_mtop, first=.true.)
                end if
             end if
             if (beam(i)%n_strfun==0) stop &
                  & 'Fixed_energy is false, but no structure function selected'
          end if
       end if
!       call set(beam(i)%rho, beam(i)%polarization)
    end do
  end subroutine beam_par_set

  subroutine read_beam_parameters_unit(unit, beam)
    integer, intent(in) :: unit
    type(beam_parameter_set), dimension(2), intent(inout) :: beam
    integer :: i
    real(kind=default), dimension(3) :: polarization
    logical :: fixed_energy
    logical :: CIRCE_on, CIRCE_map
    real(kind=default) :: CIRCE_beta, CIRCE_gamma
    real(kind=default) :: CIRCE_sqrts
    integer :: CIRCE_prt, CIRCE_ver, CIRCE_rev, CIRCE_acc, CIRCE_chat
    logical :: ISR_on
    real(kind=default) :: ISR_alpha, ISR_m_in, ISR_sqrts
    real(kind=default) :: ISR_power, ISR_x0
    logical :: EPA_on, EPA_map
    integer :: EPA_in_prt 
    real(kind=default) :: EPA_alpha, EPA_mX, EPA_Q_max, EPA_x0, EPA_x1
    real(kind=default) :: EPA_msq, EPA_s, EPA_c0, EPA_c1
    logical :: PDF_on
    integer :: PDF_in_prt, PDF_ngroup, PDF_nset, PDF_nfl, PDF_lo
    logical :: PDF_running_scale
    real(kind=default) :: PDF_scale, PDF_QCDL4, PDF_mtop
    namelist /beam_input/ polarization, fixed_energy
    namelist /beam_input/ CIRCE_on, CIRCE_map, &
    !     &                CIRCE_beta, CIRCE_gamma, &
         &                CIRCE_sqrts, CIRCE_ver, &
         &                CIRCE_rev, CIRCE_acc, CIRCE_chat
    namelist /beam_input/ ISR_on, ISR_alpha, ISR_m_in, ISR_sqrts
    namelist /beam_input/ EPA_on, EPA_map, EPA_in_prt
    namelist /beam_input/ EPA_alpha, EPA_mX
    namelist /beam_input/ EPA_Q_max, EPA_x0, EPA_x1
    namelist /beam_input/ PDF_on, PDF_in_prt, PDF_ngroup, PDF_nset, PDF_nfl, PDF_lo
    namelist /beam_input/ PDF_running_scale, PDF_scale, PDF_QCDL4
    do i=1,2
       polarization = beam(i)%polarization
       fixed_energy = beam(i)%fixed_energy
       CIRCE_on =    beam(i)%CIRCE_on
       CIRCE_map =   beam(i)%CIRCE_map
       !CIRCE_beta =  beam(i)%CIRCE_beta
       !CIRCE_gamma = beam(i)%CIRCE_gamma
       CIRCE_sqrts = beam(i)%CIRCE_sqrts
       CIRCE_ver =   beam(i)%CIRCE_ver
       CIRCE_rev =   beam(i)%CIRCE_rev
       CIRCE_acc =   beam(i)%CIRCE_acc
       CIRCE_chat =  beam(i)%CIRCE_chat
       ISR_on = beam(i)%ISR_on
       ISR_alpha = beam(i)%ISR_alpha
       ISR_m_in = beam(i)%ISR_m_in
       ISR_sqrts = beam(i)%ISR_sqrts
       EPA_on = beam(i)%EPA_on
       EPA_map = beam(i)%EPA_map
       EPA_in_prt = beam(i)%EPA_in_prt
       EPA_alpha = beam(i)%EPA_alpha
       EPA_mX = beam(i)%EPA_mX
       EPA_Q_max = beam(i)%EPA_Q_max
       EPA_x0 = beam(i)%EPA_x0
       EPA_x1 = beam(i)%EPA_x1
       PDF_on = beam(i)%PDF_on
       PDF_in_prt = beam(i)%PDF_in_prt
       PDF_ngroup = beam(i)%PDF_ngroup
       PDF_nset = beam(i)%PDF_nset
       PDF_nfl = beam(i)%PDF_nfl
       PDF_lo = beam(i)%PDF_lo
       PDF_running_scale = beam(i)%PDF_running_scale
       PDF_scale = beam(i)%PDF_scale
       PDF_QCDL4 = beam(i)%PDF_QCDL4
       read(unit, nml=beam_input)
       beam(i)%polarization = polarization
       beam(i)%fixed_energy = fixed_energy
       beam(i)%CIRCE_on =    CIRCE_on
       beam(i)%CIRCE_map =   CIRCE_map
       !beam(i)%CIRCE_beta =  CIRCE_beta
       !beam(i)%CIRCE_gamma = CIRCE_gamma
       beam(i)%CIRCE_sqrts = CIRCE_sqrts
       beam(i)%CIRCE_ver =   CIRCE_ver
       beam(i)%CIRCE_rev =   CIRCE_rev
       beam(i)%CIRCE_acc =   CIRCE_acc
       beam(i)%CIRCE_chat =  CIRCE_chat
       beam(i)%ISR_on = ISR_on
       beam(i)%ISR_alpha = ISR_alpha
       beam(i)%ISR_m_in = ISR_m_in
       beam(i)%ISR_sqrts = ISR_sqrts
       beam(i)%EPA_on = EPA_on
       beam(i)%EPA_map = EPA_map
       beam(i)%EPA_in_prt = EPA_in_prt
       beam(i)%EPA_alpha = EPA_alpha
       beam(i)%EPA_mX = EPA_mX
       beam(i)%EPA_Q_max = EPA_Q_max
       beam(i)%EPA_x0 = EPA_x0
       beam(i)%EPA_x1 = EPA_x1
       beam(i)%PDF_on = PDF_on
       beam(i)%PDF_in_prt = PDF_in_prt
       beam(i)%PDF_ngroup = PDF_ngroup
       beam(i)%PDF_nset = PDF_nset
       beam(i)%PDF_nfl = PDF_nfl
       beam(i)%PDF_lo = PDF_lo
       beam(i)%PDF_running_scale = PDF_running_scale 
       beam(i)%PDF_scale = PDF_scale 
       beam(i)%PDF_QCDL4 = PDF_QCDL4
    end do
  end subroutine read_beam_parameters_unit

  subroutine write_beam_parameters_unit(unit, beam)
    integer, intent(in) :: unit
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    integer :: i
    logical :: fixed_energy
    logical :: CIRCE_on, CIRCE_map
    real(kind=default) :: CIRCE_beta, CIRCE_gamma
    real(kind=default) :: CIRCE_sqrts
    integer :: CIRCE_prt, CIRCE_ver, CIRCE_rev, CIRCE_acc, CIRCE_chat
    logical :: ISR_on
    real(kind=default) :: ISR_alpha, ISR_m_in, ISR_sqrts
    real(kind=default) :: ISR_power, ISR_x0
    logical :: EPA_on, EPA_map
    integer :: EPA_in_prt 
    real(kind=default) :: EPA_alpha, EPA_mX, EPA_Q_max, EPA_x0, EPA_x1
    real(kind=default) :: EPA_msq, EPA_s, EPA_c0, EPA_c1
    logical :: PDF_on
    integer :: PDF_in_prt, PDF_ngroup, PDF_nset, PDF_nfl, PDF_lo
    logical :: PDF_running_scale
    real(kind=default) :: PDF_scale, PDF_QCDL4, PDF_mtop
    namelist /beam_input/ fixed_energy
    namelist /beam_input/ CIRCE_on, CIRCE_map, &
    !     &                CIRCE_beta, CIRCE_gamma, &
         &                CIRCE_sqrts, CIRCE_ver, &
         &                CIRCE_rev, CIRCE_acc, CIRCE_chat
    namelist /beam_input/ ISR_on, ISR_alpha, ISR_m_in, ISR_sqrts
    namelist /beam_input/ EPA_on, EPA_map, EPA_in_prt
    namelist /beam_input/ EPA_alpha, EPA_mX
    namelist /beam_input/ EPA_Q_max, EPA_x0, EPA_x1
    namelist /beam_input/ PDF_on, PDF_in_prt, PDF_ngroup, PDF_nset, PDF_nfl, PDF_lo
    namelist /beam_input/ PDF_running_scale, PDF_scale, PDF_QCDL4
    do i=1,2
       write(unit, *) '&beam_input'
       write(unit, *) 'polarization = '
       write(unit, *) beam(i)%polarization
       write(unit, *) 'fixed_energy = ', beam(i)%fixed_energy
       write(unit, *) 'CIRCE_on    = ', beam(i)%CIRCE_on
       write(unit, *) 'CIRCE_map   = ', beam(i)%CIRCE_map
       !write(unit, *) 'CIRCE_beta  = ', beam(i)%CIRCE_beta
       !write(unit, *) 'CIRCE_gamma = ', beam(i)%CIRCE_gamma
       write(unit, *) 'CIRCE_sqrts = ', beam(i)%CIRCE_sqrts
       write(unit, *) 'CIRCE_ver   = ', beam(i)%CIRCE_ver
       write(unit, *) 'CIRCE_rev   = ', beam(i)%CIRCE_rev
       write(unit, *) 'CIRCE_acc   = ', beam(i)%CIRCE_acc
       write(unit, *) 'CIRCE_chat  = ', beam(i)%CIRCE_chat
       write(unit, *) 'ISR_on       = ', beam(i)%ISR_on
       write(unit, *) 'ISR_alpha    = ', beam(i)%ISR_alpha
       write(unit, *) 'ISR_m_in     = ', beam(i)%ISR_m_in
       write(unit, *) 'ISR_sqrts    = ', beam(i)%ISR_sqrts
       write(unit, *) 'EPA_on       = ', beam(i)%EPA_on
       write(unit, *) 'EPA_map      = ', beam(i)%EPA_map
       write(unit, *) 'EPA_in_prt   = ', beam(i)%EPA_in_prt
       write(unit, *) 'EPA_alpha    = ', beam(i)%EPA_alpha
       write(unit, *) 'EPA_mX       = ', beam(i)%EPA_mX
       write(unit, *) 'EPA_Q_max    = ', beam(i)%EPA_Q_max
       write(unit, *) 'EPA_x0       = ', beam(i)%EPA_x0  
       write(unit, *) 'EPA_x1       = ', beam(i)%EPA_x1
       write(unit, *) 'PDF_on            = ', beam(i)%PDF_on
       write(unit, *) 'PDF_in_prt        = ', beam(i)%PDF_in_prt
       write(unit, *) 'PDF_ngroup        = ', beam(i)%PDF_ngroup
       write(unit, *) 'PDF_nset          = ', beam(i)%PDF_nset
       write(unit, *) 'PDF_nfl           = ', beam(i)%PDF_nfl
       write(unit, *) 'PDF_lo            = ', beam(i)%PDF_lo
       write(unit, *) 'PDF_running_scale = ', beam(i)%PDF_running_scale
       write(unit, *) 'PDF_scale         = ', beam(i)%PDF_scale
       write(unit, *) 'PDF_QCDL4         = ', beam(i)%PDF_QCDL4
       write(unit, *) '/'
    end do
  end subroutine write_beam_parameters_unit

  subroutine read_beam_parameters_name(name, beam)
    character(len=*), intent(in) :: name
    type(beam_parameter_set), dimension(2), intent(inout) :: beam
    integer :: unit
    unit = free_unit ()
    open (unit=unit, action='read', status='old', file=name)
    call read_beam_parameters_unit (unit, beam)
    close (unit=unit)
  end subroutine read_beam_parameters_name

  subroutine write_beam_parameters_name(name, beam)
    character(len=*), intent(in) :: name
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    integer :: unit
    unit = free_unit ()
    open (unit=unit, action='write', status='replace', file=name)
    call write_beam_parameters_unit (unit, beam)
    close (unit=unit)
  end subroutine write_beam_parameters_name

  subroutine beam_par_broadcast (beam, root, domain, error)
    type(beam_parameter_set), dimension(2), intent(inout) :: beam
    integer, intent(in) :: root
    integer, intent(in), optional :: domain
    integer, intent(out), optional :: error
    integer :: i
    do i=1,2
       call mpi90_broadcast(beam(i)%polarization, root, domain, error)
       call mpi90_broadcast(beam(i)%fixed_energy, root, domain, error)
       call mpi90_broadcast(beam(i)%CIRCE_on, root, domain, error)
       call mpi90_broadcast(beam(i)%CIRCE_map, root, domain, error)
       !call mpi90_broadcast(beam(i)%CIRCE_beta, root, domain, error)
       !call mpi90_broadcast(beam(i)%CIRCE_gamma, root, domain, error)
       call mpi90_broadcast(beam(i)%CIRCE_sqrts, root, domain, error)
       call mpi90_broadcast(beam(i)%CIRCE_ver, root, domain, error)
       call mpi90_broadcast(beam(i)%CIRCE_rev, root, domain, error)
       call mpi90_broadcast(beam(i)%CIRCE_acc, root, domain, error)
       call mpi90_broadcast(beam(i)%CIRCE_chat, root, domain, error)
       call mpi90_broadcast(beam(i)%ISR_on, root, domain, error)
       call mpi90_broadcast(beam(i)%ISR_alpha, root, domain, error)
       call mpi90_broadcast(beam(i)%ISR_m_in, root, domain, error)
       call mpi90_broadcast(beam(i)%ISR_sqrts, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_on, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_map, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_in_prt, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_alpha, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_mX, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_Q_max, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_x0, root, domain, error)
       call mpi90_broadcast(beam(i)%EPA_x1, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_on, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_in_prt, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_ngroup, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_nset, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_nfl, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_lo, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_running_scale, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_scale, root, domain, error)
       call mpi90_broadcast(beam(i)%PDF_QCDL4, root, domain, error)
    end do
  end subroutine beam_par_broadcast

  subroutine structure_function (factor, x, r, sqrts, beam, code, ok)
    real(kind=default), intent(out) :: factor
    real(kind=default), dimension(2), intent(out) :: x
    real(kind=default), dimension(:), intent(in) :: r
    real(kind=default), intent(in) :: sqrts
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    integer, dimension(2), intent(in) :: code
    logical, intent(out) :: ok
    real(kind=default), dimension(2) :: xtmp
    real(kind=default) :: scale
    integer :: i, k
    ok = .true.
    k = 1
    factor = 1
    x = 1
    if (any(beam%CIRCE_on)) then
       if (all(beam%CIRCE_on)) then
          xtmp = r(k:k+1);  k=k+2
          if (beam(1)%CIRCE_map) then
             call strfun_CIRCE_map (factor, xtmp(1), xtmp(2), &
                  &                 beam(1)%CIRCE_beta, beam(1)%CIRCE_gamma, &
                  &                 beam(2)%CIRCE_beta, beam(2)%CIRCE_gamma, &
                  &                 beam(1)%CIRCE_prt, beam(2)%CIRCE_prt)
          else
             call strfun_CIRCE (factor, xtmp(1), xtmp(2), &
                  &                 beam(1)%CIRCE_prt, beam(2)%CIRCE_prt)
          end if
          x = x * xtmp
       else 
          stop "With beamstrahlung, CIRCE must be switched on for both beams"
       end if
    end if
    do i=1, 2
       if (beam(i)%ISR_on) then
          xtmp(i) = r(k);  k=k+1
          call strfun_ISR(xtmp(i), beam(i)%ISR_x0, beam(i)%ISR_power)
          x(i) = x(i) * xtmp(i)
       end if
       if (beam(i)%EPA_on) then
          xtmp(i) = r(k);  k=k+1
          call strfun_EPA(xtmp(i), factor, beam(i)%EPA_map, &
               &          beam(i)%EPA_x0, beam(i)%EPA_x1, &
               &          beam(i)%EPA_c0, beam(i)%EPA_c1, &
               &          beam(i)%EPA_msq, &
               &          beam(i)%EPA_Q_max**2, beam(i)%EPA_s, &
               &          ok)
          x(i) = x(i) * xtmp(i)
       end if
       if (beam(i)%PDF_on) then
          x(i) = r(k);  k=k+1
          if (beam(1)%PDF_in_prt /= beam(2)%PDF_in_prt) then
             call pdflib_init (beam(i)%PDF_in_prt, beam(i)%PDF_ngroup, &
                  &            beam(i)%PDF_nset, beam(1)%PDF_nfl, beam(1)%PDF_lo, &
                  &            beam(1)%PDF_QCDL4, beam(2)%PDF_mtop)
          end if
          if (beam(i)%PDF_running_scale) then
             scale = sqrts * sqrt(x(1)*x(2))
          else
             scale = beam(i)%PDF_scale
          end if
          factor = factor * pdflib_strfun (x(i), scale, code(i))
       end if
    end do
  end subroutine structure_function

  subroutine strfun_CIRCE (f, x1, x2, p1, p2)
    real(kind=default), intent(inout) :: f, x1, x2
    integer, intent(in) :: p1, p2
    double precision :: kirke
    external kirke
    f = f * kirke (dble(x1), dble(x2), p1, p2)
  end subroutine strfun_CIRCE

  function map_CIRCE_jacobian (x, x0, a, power, invert) result (dy_dx)
    real(kind=default), intent(in) :: x, x0, a, power
    logical, intent(in) :: invert
    real(kind=default) :: dy_dx
    if (x<=0) then
       dy_dx = a
    else if (invert) then
       if (x<x0) then
          dy_dx = a * (1-x)**power
       else
          dy_dx = a * (1-x0)**power
       end if
    else
       if (x<x0) then
          dy_dx = a * x**power
       else
          dy_dx = a * x0**power
       end if
    end if
  end function map_CIRCE_jacobian

  function map_CIRCE_x (y, y0, x0, a, power, invert) result (x)
    real(kind=default), intent(in) :: y, y0, x0, a, power
    logical, intent(in) :: invert
    real(kind=default) :: x
    if (y<=0) then
       x = 0
    else if (y>1) then
       x = 1
    else if (invert) then
       if (y<y0) then
          x = 1 - (1 - (1+power)*y/a)**(1/(1+power))
       else
          x = x0 + (y-y0)/(a*(1-x0)**power)
       end if
    else
       if (y<y0) then
          x = ((1+power)*y/a)**(1/(1+power))
       else
          x = x0 + (y-y0)/(a*x0**power)
       end if
    end if
  end function map_CIRCE_x

  subroutine map_CIRCE_set_const (a, y0, x0, power, invert)
    real(kind=default), intent(out) :: a, y0
    real(kind=default), intent(in) :: x0, power
    logical, intent(in) :: invert
    real(kind=default) :: tmp
    if (invert) then
       tmp = (1 - (1-x0)**(1+power)) / (1+power)
       a = 1 / (tmp + (1-x0)**(1+power))
    else
       tmp = x0**(1+power) / (1+power)
       a = 1 / (tmp + (1-x0) * x0**power)
    end if
    y0 = a * tmp
  end subroutine map_CIRCE_set_const

  subroutine strfun_CIRCE_map_s (f, x, x0, beta, gamma)
    real(kind=default), intent(inout) :: f, x
    real(kind=default), intent(in) :: x0, beta, gamma
    real(kind=default) :: y, z, ay, az, y0, z0
    call map_CIRCE_set_const (ay, y0, x0, gamma, invert=.true.)
    call map_CIRCE_set_const (az, z0, y0, beta, invert=.false.)
    z = x
    y = map_CIRCE_x (z, z0, y0, az, beta, invert=.false.)
!    y = z
    x = map_CIRCE_x (y, y0, x0, ay, gamma, invert=.true.)
    f = f / map_CIRCE_jacobian (x, x0, ay, gamma, invert=.true.)
    f = f / map_CIRCE_jacobian (y, y0, az, beta, invert=.false.)
  end subroutine strfun_CIRCE_map_s
    
  subroutine strfun_CIRCE_map_p (f, x, eps, beta, gamma, p)
    real(kind=default), intent(inout) :: f, x
    real(kind=default), intent(in) :: eps, beta, gamma
    integer, intent(in) :: p
    real(kind=default) :: xtmp
    if (p==PHOTON) then
       xtmp = 1 - x
       call strfun_CIRCE_map_s (f, xtmp, 1-eps, beta, gamma)
       x = 1 - xtmp
    else
       call strfun_CIRCE_map_s (f, x, 1-eps, beta, gamma)
    end if
  end subroutine strfun_CIRCE_map_p

  subroutine map_CIRCE_set_powers (beta, gamma, p)
    real(kind=default), intent(out) :: beta, gamma
    integer, intent(in) :: p
    double precision x1m, x2m, roots
    common /circom/ x1m, x2m, roots
    double precision lumi
    common /circom/ lumi
    double precision a1(0:7)
    common /circom/ a1
    double precision elect0, gamma0
    common /circom/ elect0, gamma0
    integer acc, ver, rev, chat
    common /circom/ acc, ver, rev, chat
    integer magic
    common /circom/ magic
    integer e, r, ehi, elo
    common /circom/ e, r, ehi, elo
    save /circom/
    if (abs(p)==ELECTRON) then
       beta = a1(2)
       gamma = a1(3)
    else if (abs(p)==PHOTON) then
       beta = a1(6)
       gamma = a1(5)
    end if
  end subroutine map_CIRCE_set_powers

  subroutine strfun_CIRCE_map (f, x1, x2, beta1, gamma1, beta2, gamma2, p1, p2)
    real(kind=default), intent(inout) :: f, x1, x2
    real(kind=default), intent(in) :: beta1, gamma1
    real(kind=default), intent(in) :: beta2, gamma2
    integer, intent(in) :: p1, p2
    real(kind=default), parameter :: eps = KIREPS
    call strfun_CIRCE_map_p (f, x1, eps, beta1, gamma1, p1)
    call strfun_CIRCE_map_p (f, x2, eps, beta2, gamma2, p2)
    call strfun_CIRCE (f, x1, x2, p1, p2)
  end subroutine strfun_CIRCE_map

  subroutine strfun_ISR(x, x0, power)
    real(kind=default), intent(inout) :: x
    real(kind=default), intent(in) :: x0, power
    if (x>=x0) then
       x = 1
    else if (x<=0) then
       x = tiny(1._default)
    else
       x = 1 - (1-x)**power
    end if
  end subroutine strfun_ISR

  subroutine strfun_EPA(x, f, map, x0, x1, c0, c1, msq, Qmaxsq, s, ok)
    real(kind=default), intent(inout) :: x, f
    logical, intent(in) :: map
    real(kind=default), intent(in) :: x0, x1, c0, c1, msq, Qmaxsq, s
    logical, intent(out) :: ok
    real(kind=default) :: L, D, lx, lx0, lx1, Qminsq
    ok = .true.
    if (map) then
       L = log(Qmaxsq/msq)
       lx0 = log(x0);  lx1 = log(x1)
       D = L**2 - 4*(x * lx1 * (L-lx1) + (1-x) * lx0 * (L-lx0))
       if (D <= 0) then
          ok = .false.; return
       end if
       lx = (L - sqrt(D))/2
       x = exp(lx)
       f = f / ( L - 2*lx)
    else
       x = x * x1 + (1-x) * x0
       f = f / x
    end if
    Qminsq = msq * x**2 / (1-x)
    if (Qminsq >= Qmaxsq) then
       ok = .false.
       return
    end if
    f = f  * (  (1-x+x**2/2)*log(Qmaxsq/Qminsq) &
         &    - (1-x/2)**2 * log((x**2 + 4*Qmaxsq/s) / (x**2 + 4*Qminsq/s)) &
         &    - (1-x) * (1 - Qminsq/Qmaxsq) ) &
         & * ( c1 - c0 ) 
  end subroutine strfun_EPA


end module beams
